import asyncio
import logging
from enum import Enum
from json import JSONDecodeError
from typing import Dict, Text, Any, Optional

import aiohttp
from aiohttp import web, WSMsgType, WSMessage
from aiohttp.web_request import Request
from aiohttp.web_ws import WebSocketResponse

from rasax.community import utils, config

logger = logging.getLogger(__name__)

CONNECTED_WEB_SOCKETS = {}  # type: Dict[Text, WebSocketResponse]
WEBSOCKET_ENDPOINT = "/ws"
DEFAULT_WEBSOCKET_URL = "ws://0.0.0.0:5003" + WEBSOCKET_ENDPOINT
BROADCAST_SENDER_ID = "ALL"

AUTHORIZATION_KEY = "Authorization"


class OperationStatus(Enum):
    """Represents results of operations in websocket notifications."""

    STARTED = 0
    SUCCESS = 1
    FAILED = 2
    CANCELED = 3


class NotificationService:
    def __init__(self, websocket_url: Text = DEFAULT_WEBSOCKET_URL):
        self._websocket_url = websocket_url

        # Lazy initialization of the connections
        self._session = None  # type: Optional[aiohttp.ClientSession]
        self._websocket = None  # type: Optional[aiohttp.ClientWebSocketResponse]

    def __del__(self) -> None:
        # Close everything when the object gets destroyed.
        asyncio.ensure_future(self._close_websocket())

    async def _close_websocket(self) -> None:
        if self._websocket:
            await self._websocket.close()

        if self._session:
            await self._session.close()

    async def _connect_websocket(self) -> None:
        if not self._session:
            self._session = aiohttp.ClientSession()

        if not self._websocket:
            headers = {
                "Authorization": utils.bearer_token({"username": config.SYSTEM_USER})
            }
            self._websocket = await self._session.ws_connect(
                self._websocket_url, headers=headers
            )

    async def _send_notification(self, message: Dict[Text, Any]) -> None:
        """Send a message to the websocket server.

        Args:
            message: The message.
        """

        await self._connect_websocket()
        await self._websocket.send_json(message)

    async def send_notification_to(
        self,
        recipient_id: Optional[Text],
        topic: Text,
        status: OperationStatus,
        details: Optional[Dict[Text, Any]] = None,
    ) -> None:
        """Send a notification to a Rasa X user if the user has a connected websocket.

        Args:
            recipient_id: the id of the Rasa X user who should get the notification.
            topic: the category of the message
            status: status of the operation the notification describes.
            details: optional additional information of the notification.
        """

        if not recipient_id:
            # Handling the `None` case makes it way easier for callers of the function
            # which may or may not have the recipient
            return

        message = {
            "topic": topic,
            "status": status.name,
            "details": details,
            "recipient_id": recipient_id,
        }
        await self._send_notification(message)


def run_websocket_server(websocket_port: int = 5003) -> None:
    """Runs a websocket server."""

    app = web.Application()
    app.add_routes([web.get("/ws", _websocket_handler)])
    logger.debug(
        "Starting web socket server on port '{}' with endpoint '{}'.".format(
            websocket_port, WEBSOCKET_ENDPOINT
        )
    )

    web.run_app(app, port=websocket_port)


async def _websocket_handler(request: Request) -> WebSocketResponse:
    """Handles new connections to the websocket server."""

    websocket = web.WebSocketResponse()
    await websocket.prepare(request)

    async for message in websocket:  # type: WSMessage
        if message.type == WSMsgType.TEXT:
            await _handle_text_message(_get_json(message), request, websocket)
        elif message.type == WSMsgType.ERROR:
            break
    _remove_websocket_connection(websocket)
    return websocket


async def _handle_text_message(
    message: Dict, request: Request, websocket: WebSocketResponse
) -> None:
    if _is_authentication(message):
        username = _get_username_from_token(message)
        if username:
            _add_websocket_connection(username, websocket)
        else:
            await websocket.close()

    if _is_from_rasa_x_server(request):
        recipient_id = message.get("recipient_id")
        if recipient_id == BROADCAST_SENDER_ID:
            await _forward_message_to_connected_websockets(message)
        elif recipient_id:
            await _forward_message_to(recipient_id, message)


def _is_from_rasa_x_server(request: Request) -> bool:
    localhost_aliases = ["localhost", "127.0.0.1", "::1", "0.0.0.0"]
    return any([request.host.startswith(host) for host in localhost_aliases])


def _get_json(message: WSMessage) -> Dict:
    try:
        return message.json() or {}
    except JSONDecodeError:
        return {}


async def _forward_message_to(recipient_id: Text, message: Dict[Text, Any]) -> None:
    websocket_of_recipient = CONNECTED_WEB_SOCKETS.get(recipient_id)
    if websocket_of_recipient is not None:
        await websocket_of_recipient.send_json(message)
        logger.debug("Sent notification to recipient '{}'.".format(recipient_id))
    else:
        logger.debug(
            "Web socket connection for recipient '{}' not found.".format(recipient_id)
        )


async def _forward_message_to_connected_websockets(message: Dict[Text, Any]) -> None:
    logger.debug("Broadcasting this message: {}".format(message))
    send_message_coroutines = [
        w.send_json(message) for w in CONNECTED_WEB_SOCKETS.values()
    ]
    await asyncio.gather(*send_message_coroutines)


def _get_username_from_token(message: Dict) -> Optional[Text]:
    try:
        token = message.get(AUTHORIZATION_KEY)
        token_payload = utils.verify_jwt_token(token)
        return token_payload.get("username")
    except ValueError:
        return None


def _add_websocket_connection(sender_id: Text, websocket: WebSocketResponse) -> None:
    CONNECTED_WEB_SOCKETS[sender_id] = websocket
    logger.debug("New websocket connection with '{}'.".format(sender_id))


def _remove_websocket_connection(websocket_to_delete: WebSocketResponse) -> None:
    matching_keys = [
        sender_id
        for sender_id, socket in CONNECTED_WEB_SOCKETS.items()
        if socket == websocket_to_delete
    ]
    for sender_id in matching_keys:
        del CONNECTED_WEB_SOCKETS[sender_id]
        logger.debug(
            "Removed websocket for '{}' from connected websockets.".format(sender_id)
        )


def _is_authentication(message: Dict) -> bool:
    return AUTHORIZATION_KEY in message


def _is_authenticated(websocket: WebSocketResponse) -> bool:
    return websocket in CONNECTED_WEB_SOCKETS.values()
